;;; howm-reminder.el --- Wiki-like keyword jump between files
;;; Copyright (c) 2002, 2003, 2004
;;;   by HIRAOKA Kazuyuki <khi@users.sourceforge.jp>
;;; $Id: howm-reminder.el,v 1.30 2004/08/13 22:37:24 hira Exp $
;;;
;;; This program is free software; you can redistribute it and/or modify
;;; it under the terms of the GNU General Public License as published by
;;; the Free Software Foundation; either version 1, or (at your option)
;;; any later version.
;;;
;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.
;;;
;;; The GNU General Public License is available by anonymouse ftp from
;;; prep.ai.mit.edu in pub/gnu/COPYING.  Alternately, you can write to
;;; the Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
;;; USA.
;;--------------------------------------------------------------------

(require 'howm-common)

(defvar howm-list-schedule-name "{schedule}")
(defvar howm-list-todo-name "{todo}")
;   "This is used for buffer name of `howm-list-reminder'.
; See `howm-view-summary-name'.")

(defvar howm-reminder-normal-face   'howm-reminder-normal-face)
(defvar howm-reminder-todo-face     'howm-reminder-todo-face)
(defvar howm-reminder-defer-face 'howm-reminder-defer-face)
(defvar howm-reminder-deadline-face 'howm-reminder-deadline-face)
(defvar howm-reminder-schedule-face 'howm-reminder-schedule-face)
(defvar howm-reminder-done-face     'howm-reminder-done-face)
(defvar howm-reminder-today-face     'howm-reminder-today-face)
(defvar howm-reminder-tomorrow-face 'howm-reminder-tomorrow-face)

(defface howm-reminder-normal-face
  '((t (:foreground "blue")))
  "Face for `howm-list-reminder'.")
(defface howm-reminder-todo-face
  '((((class color) (background light)) (:foreground "purple"))
    (((class color) (background dark)) (:foreground "yellow")))
  "Face for `howm-list-reminder'.")
(defface howm-reminder-defer-face
  '((t (:foreground "magenta")))
  "Face for `howm-list-reminder'.")
(defface howm-reminder-deadline-face
  '((t (:foreground "red")))
  "Face for `howm-list-reminder'.")
(defface howm-reminder-schedule-face
  '((((class color) (background light)) (:foreground "dark green"))
    (((class color) (background dark)) (:foreground "green")))
  "Face for `howm-list-reminder'.")
(defface howm-reminder-done-face
  '((((class color) (background light)) (:foreground "black"))
    (((class color) (background dark)) (:foreground "gray")))
  "Face for `howm-list-reminder'.")
(defface howm-reminder-today-face
  '((t (:foreground "black" :background "orange")))
  "Face for `howm-list-reminder'.")
(defface howm-reminder-tomorrow-face
  '((t (:foreground "black" :background "pink")))
  "Face for `howm-list-reminder'.")

(defvar howm-todo-priority-func
      '(("-" . howm-todo-priority-normal)
        (" " . howm-todo-priority-normal)
        ("+" . howm-todo-priority-todo)
        ("~" . howm-todo-priority-defer)
        ("!" . howm-todo-priority-deadline)
        ("@" . howm-todo-priority-schedule)
        ("." . howm-todo-priority-done)))
(defvar howm-todo-priority-normal-laziness 1)
(defvar howm-todo-priority-todo-laziness 7)
(defvar howm-todo-priority-todo-init -7)
(defvar howm-todo-priority-defer-laziness 30)
(defvar howm-todo-priority-defer-init -14)
(defvar howm-todo-priority-defer-peak 0)
(defvar howm-todo-priority-deadline-laziness 7)
(defvar howm-todo-priority-deadline-init -2)
(defvar howm-todo-priority-schedule-laziness 1)
(defvar howm-huge- 66666)
(defvar howm-huge 77777)
(defvar howm-huge+ 88888)
(defvar howm-todo-priority-normal-bottom   (- howm-huge)) 
(defvar howm-todo-priority-todo-bottom     (- howm-huge)) 
(defvar howm-todo-priority-defer-bottom    (- howm-huge)) 
(defvar howm-todo-priority-deadline-bottom (- howm-huge)) 
(defvar howm-todo-priority-deadline-top    howm-huge)              
(defvar howm-todo-priority-schedule-bottom (- howm-huge)) 
(defvar howm-todo-priority-done-bottom     (- howm-huge+))
(defvar howm-todo-priority-unknown-top     howm-huge+)             

(defvar howm-action-lock-reminder-done-default nil)

(defvar howm-congrats-count 0)
(defvar howm-congrats-format
      '("Finished %s tasks!"))
(defvar howm-congrats-hook nil)

(defvar howm-reminder-cancel-string "cancel")

;;; --- level ? ---

;; Fix me: redundant (howm-date-* & howm-reminder-*)

;; (defun howm-reminder-regexp-grep (types)
;;   (howm-inhibit-warning-in-compilation))
;; (defun howm-reminder-regexp (types)
;;   (howm-inhibit-warning-in-compilation))

(defvar howm-reminder-old-format nil)
(if howm-reminder-old-format
    (progn ;; old format
      (defvar howm-reminder-regexp-grep-format
        "@\\[[0-9][0-9][0-9][0-9]/[0-9][0-9]/[0-9][0-9]\\]%s")
      (defvar howm-reminder-regexp-format
        "\\(@\\)\\[\\([0-9][0-9][0-9][0-9]\\)/\\([0-9][0-9]\\)/\\([0-9][0-9]\\)\\]\\(%s\\)\\([0-9]*\\)")
      (defun howm-reminder-regexp-grep (types)
        (format howm-reminder-regexp-grep-format types))
      (defun howm-reminder-regexp (types)
        (format howm-reminder-regexp-format types))
      (defvar howm-reminder-regexp-command-pos 1)
      (defvar howm-reminder-regexp-year-pos 2)
      (defvar howm-reminder-regexp-month-pos 3)
      (defvar howm-reminder-regexp-day-pos 4)
      (defvar howm-reminder-regexp-type-pos 5)
      (defvar howm-reminder-regexp-laziness-pos 6)
      (defvar howm-reminder-today-format "@[%Y/%m/%d]")
      (defvar howm-reminder-font-lock-keywords
        `(
          (,(howm-reminder-regexp "[-]?") (0 howm-reminder-normal-face prepend))
          (,(howm-reminder-regexp "[+]") (0 howm-reminder-todo-face prepend))
          (,(howm-reminder-regexp "[~]") (0 howm-reminder-defer-face prepend))
          (,(howm-reminder-regexp "[!]") (0 howm-reminder-deadline-face prepend))
          (,(howm-reminder-regexp "[@]") (0 howm-reminder-schedule-face prepend))
          (,(howm-reminder-regexp "[.]") (0 howm-reminder-done-face prepend))
          ))
      (defun howm-reminder-font-lock-keywords ()
        howm-reminder-font-lock-keywords)
      (defvar howm-reminder-types "[-+~!@. ]")
      (defvar howm-schedule-types "[!@.]")
      (defvar howm-schedule-menu-types "[!@]")
      (defvar howm-todo-types "[-+~!. ]")
      (defvar howm-todo-menu-types "[-+~! ]")
      (defun howm-action-lock-done (&optional command)
        (save-excursion
          (let ((at-beg (match-beginning howm-reminder-regexp-command-pos))
                (at-end (match-end  howm-reminder-regexp-command-pos))
                (type-beg (match-beginning howm-reminder-regexp-type-pos))
                (type-end (match-end howm-reminder-regexp-type-pos))
                (lazy-beg (match-beginning howm-reminder-regexp-laziness-pos))
                (lazy-end (match-end howm-reminder-regexp-laziness-pos)))
            (let* ((s (or command
                          (read-from-minibuffer
                           "RET (done), x (cancel), symbol (type), num (laziness): ")))
                   (c (cond ((string= s "") ".")
                            ((= 0 (string-to-number s)) ". give up")
                            (t nil))))
              (when (string= s "")
                (howm-congrats))
              (if c
                  (progn
                    (goto-char at-beg)
                    (delete-region at-beg at-end)
                    (insert (howm-reminder-today))
                    (insert (format "%s " c)))
                (progn
                  (goto-char lazy-beg)
                  (delete-region lazy-beg lazy-end)
                  (when (string= (buffer-substring-no-properties type-beg type-end)
                                 " ")
                    (goto-char type-beg)
                    (insert "-")) ;; "no type" = "normal"
                  (insert s)))))))
      )
  (progn ;; new format
    (defvar howm-reminder-regexp-grep-format
      (concat "\\[" howm-date-regexp-grep "[ :0-9]*\\]%s"))
    (defvar howm-reminder-regexp-format
      (concat "\\[" howm-date-regexp "[ :0-9]*\\]\\(\\(%s\\)\\([0-9]*\\)\\)"))
;;     (defvar howm-reminder-regexp-grep-format
;;       (concat "\\[" howm-date-regexp-grep "\\]%s"))
;;     (defvar howm-reminder-regexp-format
;;       (concat "\\[" howm-date-regexp "\\]\\(\\(%s\\)\\([0-9]*\\)\\)"))
    (defun howm-reminder-regexp-grep (types)
      (format howm-reminder-regexp-grep-format types))
    (defun howm-reminder-regexp (types)
      (format howm-reminder-regexp-format types))
    (defvar howm-reminder-regexp-year-pos  howm-date-regexp-year-pos)
    (defvar howm-reminder-regexp-month-pos howm-date-regexp-month-pos)
    (defvar howm-reminder-regexp-day-pos   howm-date-regexp-day-pos)
    (defvar howm-reminder-regexp-command-pos 4)
    (defvar howm-reminder-regexp-type-pos 5)
    (defvar howm-reminder-regexp-laziness-pos 6)
    (defvar howm-reminder-today-format
      (format howm-insert-date-format howm-date-format))
    (defvar howm-reminder-font-lock-keywords
      `(
        (,(howm-reminder-regexp "[-]") (0 howm-reminder-normal-face prepend))
        (,(howm-reminder-regexp "[+]") (0 howm-reminder-todo-face prepend))
        (,(howm-reminder-regexp "[~]") (0 howm-reminder-defer-face prepend))
        (,(howm-reminder-regexp "[!]") (0 howm-reminder-deadline-face prepend))
        (,(howm-reminder-regexp "[@]") (0 howm-reminder-schedule-face prepend))
        (,(howm-reminder-regexp "[.]") (0 howm-reminder-done-face prepend))
        ))
    (defun howm-reminder-font-lock-keywords ()
      howm-reminder-font-lock-keywords)
    (defvar howm-reminder-types "[-+~!@.]")
    (defvar howm-schedule-types "[!@.]")
    (defvar howm-schedule-menu-types "[!@]")
    (defvar howm-todo-types "[-+~!.]")
    (defvar howm-todo-menu-types "[-+~!]")
    (defun howm-action-lock-done-prompt ()
      (format "RET (done), x (%s), symbol (type), num (laziness): "
              howm-reminder-cancel-string))
    (defun howm-action-lock-done (&optional command)
      (let ((whole-beg (match-beginning 0))
            (com-beg (match-beginning howm-reminder-regexp-command-pos))
            (com-end (match-end  howm-reminder-regexp-command-pos))
            (type-beg (match-beginning howm-reminder-regexp-type-pos))
            (type-end (match-end howm-reminder-regexp-type-pos))
            (lazy-beg (match-beginning howm-reminder-regexp-laziness-pos))
            (lazy-end (match-end howm-reminder-regexp-laziness-pos)))
          (let* ((s (or command
                        (howm-read-string (howm-action-lock-done-prompt)
                                          "x-+~!.@"
                                          "0123456789")))
;;                         (read-from-minibuffer (howm-action-lock-done-prompt) )))
                 (c (cond ((string= s "") (progn (howm-congrats) "."))
                          ((string= s "x") (format ". %s"
                                                   howm-reminder-cancel-string))
                          (t nil)))
                 (type-or-lazy (string-match (format "^\\(%s?\\)\\([0-9]*\\)$"
                                                     howm-reminder-types)
                                             s)))
            (cond (c
                   (progn
                     (goto-char com-beg)
                     (insert ":")
;;                      (goto-char com-end)
;;                      (insert ")")
;;                      (goto-char com-beg)
;;                      (insert "(")
                     (goto-char whole-beg)
                     (insert (howm-reminder-today))
                     (insert (format "%s " c))))
                  (type-or-lazy
                   (let* ((ty (match-string-no-properties 1 s))
                          (lz (match-string-no-properties 2 s))
                          (orig-ty (buffer-substring-no-properties type-beg
                                                                   type-end))
                          (orig-lz (buffer-substring-no-properties lazy-beg
                                                                   lazy-end))
                          (new-ty (if (string= ty "") orig-ty ty))
                          (new-lz (if (string= lz "") orig-lz lz)))
                     (goto-char com-beg)
                     (delete-region com-beg com-end)
                     (insert new-ty new-lz)))
                  (t (error "Can't understand %s" s)))
            (goto-char com-beg))))
    ))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Reminder: schedule & todo

(define-key howm-view-summary-mode-map "." 'howm-reminder-goto-today)

(defun howm-list-schedule ()
  (interactive)
  (let ((howm-view-summary-format (if howm-view-split-horizontally;; dirty!
                                      ""
                                    howm-view-summary-format)))
    (howm-list-reminder howm-schedule-types howm-list-schedule-name)
    (howm-schedule-sort-by-date)
    (howm-reminder-goto-today)))

;; (defun howm-list-schedule ()
;;   (interactive)
;;   (let ((howm-view-summary-format (if howm-view-split-horizontally;; dirty!
;;                                       ""
;;                                     howm-view-summary-format)))
;;     (howm-list-reminder howm-schedule-types howm-list-schedule-name)
;;     (howm-view-sort-by-summary)
;;     (howm-reminder-goto-today)))

(defun howm-list-reminder (types name)
  (let* ((r (howm-reminder-regexp types))
         (rg (howm-reminder-regexp-grep types))
         (summarizer (howm-reminder-summarizer r t)))
    (howm-view-search rg (howm-reminder-search-path) name summarizer)
    (howm-list-exclude)
    (howm-reminder-add-font-lock)
    (howm-mode-add-font-lock)
    (let ((action-lock-default-rules
           (howm-action-lock-reminder-forward-rules t)))
      (action-lock-mode))))

(defun howm-reminder-add-font-lock ()
  (cheat-font-lock-append-keywords (howm-reminder-font-lock-keywords))
  (cheat-font-lock-append-keywords `((,(regexp-quote (howm-reminder-today))
                                     (0 howm-reminder-today-face prepend))))
  (cheat-font-lock-append-keywords `((,(regexp-quote (howm-reminder-today 1))
                                     (0 howm-reminder-tomorrow-face prepend))))
  )

(defun howm-reminder-omit-before (regexp str)
  (string-match regexp str)
  (substring str (match-beginning 0)))

(defun howm-reminder-summarizer (regexp &optional show-day-of-week)
  `(lambda (file line content)
     (let ((s (howm-reminder-omit-before ,regexp content)))
;;                 (string-match ,regexp content)
;;                 (substring content (match-beginning 0)))))
       ,(if show-day-of-week
            '(format "%s %s"
                     (howm-day-of-week-string (nth 4
                                                   (howm-todo-parse-string s)))
                     s)
          's))))

(defun howm-reminder-today (&optional days-after)
  (let ((ti (howm-days-after (current-time) (or days-after 0))))
    (format-time-string howm-reminder-today-format ti)))

;; dirty. peek howm-view-*
(defun howm-reminder-goto-today ()
  (interactive)
  (let* ((today (howm-reminder-today))
         (r (howm-reminder-regexp "."))
         (summaries (mapcar (lambda (item)
                              (howm-reminder-omit-before
                               r (howm-view-item-summary item)))
                           (howm-view-item-list))))
;;         (summaries (mapcar 'howm-view-item-summary (howm-view-item-list))))
    (let ((rest summaries)
          (n 0))
      (while (and rest
                  (string< (car rest) today))
        (setq rest (cdr rest)
              n (1+ n)))
      (goto-line (1+ n)))))

(defun howm-schedule-menu (days &optional days-before)
  (let* ((today (howm-reminder-today (- (or days-before 0))))
         (limit (howm-reminder-today days))
         (howm-schedule-types howm-schedule-menu-types)  ;; dirty
         (raw (howm-reminder-search howm-schedule-types))
         (filtered (remove-if #'(lambda (item)
                                  (let ((s (howm-view-item-summary item)))
                                    (or (string< s today)
                                        (string< limit s))))
                              raw)))
    (howm-view-sort-items #'howm-view-item-summary #'string<
                          filtered)))

(defun howm-schedule-sort-by-date ()
  (interactive)
  (let ((howm-view-summary-format (if howm-view-split-horizontally ;; dirty!
                                      ""
                                    howm-view-summary-format)))
    (howm-view-sort-general #'howm-schedule-date #'<)))

(defun howm-schedule-date (item)
  (car (howm-reminder-parse item)))

(defun howm-reminder-search (types)
  (let* ((r (howm-reminder-regexp types))
         (rg (howm-reminder-regexp-grep types))
         (summarizer (howm-reminder-summarizer r))
         (raw (howm-view-search-items rg (howm-reminder-search-path)
                                      summarizer)))
;;          (raw (howm-view-search-items rg howm-directory summarizer)))
    (remove-if #'howm-item-exclude-p raw)))

(defun howm-list-todo ()
  (interactive)
  (let ((howm-view-summary-format (if howm-view-split-horizontally;; dirty!
                                      ""
                                    howm-view-summary-format)))
    (howm-list-reminder howm-todo-types howm-list-todo-name)
    (howm-todo-sort-by-priority)))

(defun howm-todo-menu (n limit-priority)
  (let* ((raw (howm-reminder-search howm-todo-menu-types))
         (sorted (howm-view-sort-items #'howm-todo-priority #'> raw)))
    (if (null sorted) nil
      (remove-if (lambda (item)
                   (< (howm-todo-priority item) limit-priority))
                 (if (< (length sorted) n)
                     sorted
                   (subseq sorted 0 n))))))

(defun howm-todo-sort-by-priority ()
  (interactive)
  (let ((howm-view-summary-format (if howm-view-split-horizontally ;; dirty!
                                      ""
                                    howm-view-summary-format)))
    (howm-view-sort-general #'howm-todo-priority #'>)))

;; Clean me.
(defun howm-reminder-parse (item)
  (howm-todo-parse-string (howm-view-item-summary item)))
(defun howm-todo-parse (item)
  (cdr (howm-reminder-parse item)))
(defun howm-todo-parse-string (str)
  (let* ((summary str)
         (dummy (string-match (howm-reminder-regexp ".") summary))
;          (dummy (string-match (howm-reminder-regexp howm-todo-types) summary))
         (y (match-string-no-properties howm-reminder-regexp-year-pos summary))
         (m (match-string-no-properties howm-reminder-regexp-month-pos summary))
         (d (match-string-no-properties howm-reminder-regexp-day-pos summary))
         (ty (match-string-no-properties howm-reminder-regexp-type-pos
                                           summary))
         (lz (match-string-no-properties howm-reminder-regexp-laziness-pos
                                           summary)))
    (let* ((day (howm-encode-day d m y))
           (today (howm-encode-day))
           (late (- today day))
           (type (substring (or ty "-") 0 1))  ;; "-" for old format
           (lazy (cond ((string= type " ") nil)
                       ((null lz) nil)
                       (t (let ((z (string-to-number lz)))
                            (if (= z 0) nil z)))))
;;            (lazy (if (string= type " ")
;;                      0
;;                    (string-to-number (or lz "0"))))
           (day-of-week (nth 6
                             (decode-time (apply #'encode-time
                                                 (mapcar #'string-to-number
                                                         (list "0" "0" "0"
                                                               d m y)))))))
      (list day late type lazy day-of-week))))

(defun howm-todo-priority (item)
  (let* ((p (howm-todo-parse item))
         (late (car p))
         (type (second p))
         (lazy (third p))
         (f (or (cdr (assoc type howm-todo-priority-func))
                #'howm-todo-priority-unknown)))
    (funcall f late lazy item)))

(defun howm-todo-relative-late (late laziness default-laziness)
  (/ late (float (or laziness default-laziness))))

(defun howm-todo-priority-normal (late lz item)
  (let ((r (howm-todo-relative-late late lz
                                    howm-todo-priority-normal-laziness)))
    (cond ((< r 0) (+ r howm-todo-priority-normal-bottom))
          (t (- r)))))

(defun howm-todo-priority-todo (late lz item)
  (let ((r (howm-todo-relative-late late lz
                                    howm-todo-priority-todo-laziness))
        (c (- howm-todo-priority-todo-init)))
    (cond ((< r 0) (+ r howm-todo-priority-todo-bottom))
          (t (* c (- r 1))))))

(defun howm-todo-priority-defer (late lz item)
  (let* ((r (howm-todo-relative-late late lz
                                     howm-todo-priority-defer-laziness))
         (p howm-todo-priority-defer-peak)
         (c (- p howm-todo-priority-defer-init)))
    (let ((v (* 2 (abs (- (mod r 1) 0.5)))))
      (cond ((< r 0) (+ r howm-todo-priority-defer-bottom))
            (t (- p (* c v)))))))

(defun howm-todo-priority-deadline (late lz item)
  (let ((r (howm-todo-relative-late late lz
                                    howm-todo-priority-deadline-laziness))
        (c (- howm-todo-priority-deadline-init)))
    (cond ((> r 0) (+ r howm-todo-priority-deadline-top))
          ((< r -1) (+ r howm-todo-priority-deadline-bottom))
          (t (* c r)))))

(defun howm-todo-priority-schedule (late lz item)
  (let ((r (howm-todo-relative-late late lz
                                    howm-todo-priority-schedule-laziness)))
    (cond ((> r 0) (+ r howm-todo-priority-schedule-bottom))
          (t r))))

(defun howm-todo-priority-done (late lz item)
  (+ late howm-todo-priority-done-bottom))

(defun howm-todo-priority-unknown (late lz item)
  (+ late howm-todo-priority-unknown-top))

(defun howm-encode-day (&optional d m y)
  "Convert date Y-M-D to a float number, days from the reference date.
When D is omitted, the current time is encoded."
  (let* ((e (apply #'encode-time (if d
                                     (mapcar #'string-to-number
                                             (list "0" "0" "0" d m y))
                                   (decode-time))))
         (hi (car e))
         (low (second e))
         (daysec (* 60 60 24.0)))
    (+ (* hi (/ 65536 daysec)) (/ low daysec))))

(defvar howm-congrats-command nil)
;; (setq howm-congrats-command '("play" "~/sound/level.wav"))
(defun howm-congrats ()
  (setq howm-congrats-count (1+ howm-congrats-count))
  (let* ((n (length howm-congrats-format))
         (r (random n)))
    (message (nth r howm-congrats-format) howm-congrats-count)
    (when howm-congrats-command
      (howm-congrats-run howm-congrats-command))
    (run-hooks 'howm-congrats-hook)))
(defun howm-congrats-run (com-arg-list)
  (let* ((name "howm-congrats")
         (command (car com-arg-list))
         (args (cdr com-arg-list))
         (prev (get-process name)))
    (when prev
      (delete-process prev))
    (apply #'start-process-shell-command `(,name nil ,command ,@args))))

(defun howm-action-lock-reminder-done-rule ()
  (list (howm-reminder-regexp howm-reminder-types)
        `(lambda (&optional arg)
           (let ((command (if arg
                              nil
                            howm-action-lock-reminder-done-default)))
             (howm-action-lock-done command)))
        howm-reminder-regexp-command-pos))

(defun howm-reminder-search-path ()
  (howm-search-path t))

;;; direct manipulation of items from todo list

;; I'm sorry for dirty procedure here.
;; If we use naive howm-date-regexp, it matches to file name "2004-05-11.howm"
;; in summary mode.
(defun howm-action-lock-reminder-forward-rules (&optional summary-mode-p)
  (let* ((action-maker (lambda (pos)
                         `(lambda (&optional dummy)
                            (howm-action-lock-forward (match-beginning ,pos)))))
         (reminder-rule (list (howm-reminder-regexp howm-reminder-types)
                              (funcall action-maker 0)
                              howm-reminder-regexp-command-pos))
         (summary-date-reg (format ".*%s.*\\(%s\\)"
                                   (regexp-quote howm-view-summary-sep)
                                   howm-date-regexp))
         (summary-date-reg-pos 1)
         (summary-date-rule (list summary-date-reg
                                  (funcall action-maker summary-date-reg-pos)
                                  summary-date-reg-pos))
         (menu-date-rule (list howm-date-regexp
                               (funcall action-maker 0)))
         (date-rule (if summary-mode-p
                        summary-date-rule
                      menu-date-rule)))
    (list reminder-rule date-rule)))

(defvar howm-action-lock-forward-save-buffer nil)
(defun howm-action-lock-forward (form-pos)
  (let* ((cursor-pos (point))
         (form-reg (howm-line-tail-regexp form-pos))
         (cursor-reg (howm-line-tail-regexp cursor-pos)))
    (let ((modified-tail (howm-action-lock-forward-invoke form-reg
                                                          cursor-reg))
          (buffer-read-only nil)
          (end-pos (line-end-position)))
      (when (not (string= (buffer-substring-no-properties form-pos end-pos)
                          modified-tail))
        (delete-region form-pos (line-end-position))
        (insert modified-tail)
        (goto-char cursor-pos)
        (howm-action-lock-forward-update)))))

(defun howm-line-tail-regexp (pos)
  (concat (regexp-quote (buffer-substring-no-properties pos
                                                        (line-end-position)))
          "$"))

(defun howm-action-lock-forward-invoke (form-reg cursor-reg)
  (save-excursion
    (save-window-excursion
      (howm-action-lock-forward-open)
      ;; Now we are on the corresponding line in the target file.
      (end-of-line)
      (when (not (re-search-backward form-reg (line-beginning-position) t))
        (error "Can't find corresponding line."))
      (goto-char (match-beginning 0))
      ;; Now we are at the beginning of the form.
      ;; Remember this position to report the modified tail.
      (save-excursion
        (when (not (re-search-forward cursor-reg (line-end-position) t))
          (error "Can't find corresponding string."))
        (goto-char (match-beginning 0))
        ;; Now we are at the corresponding position.
        ;; Let's hit magic-return!
        (action-lock-invoke))
      (when howm-action-lock-forward-save-buffer
        (save-buffer))
      ;; We are back to the beginning of the form.
      ;; Report the modified tail.
      (buffer-substring-no-properties (point) (line-end-position))))
  ;; We are back to the original line in the original buffer.
  )

(defun howm-action-lock-forward-open ()
  (cond ((eq major-mode 'howm-menu-mode)
         (progn
           (beginning-of-line)
           (action-lock-magic-return)
           (when (eq major-mode 'howm-view-summary-mode)
             (error "Several candidates."))))
        ((eq major-mode 'howm-view-summary-mode)
         (howm-view-summary-open))
        (t
         (error "Not supported on this buffer."))))

(defun howm-action-lock-forward-update ()
  (cond ((eq major-mode 'howm-menu-mode)
         nil) ; do nothing
        ((eq major-mode 'howm-view-summary-mode)
         (howm-view-summary-check t))
        (t
         (error "Not supported on this buffer."))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; customize

(defun howm-define-reminder (letter priority-func face schedule todo)
  "Define reminder type LETTER whose priority is determined by PRIORITY-FUNC.
It appears with FACE in schedule list when SCHEDULE is non-nil,
and in todo list when TODO is non-nil.
It also appears in menu if SCHEDULE or TODO is t."
  (add-to-list 'howm-todo-priority-func
               (cons letter priority-func))
  (add-to-list 'howm-reminder-font-lock-keywords
               `(,(howm-reminder-regexp (format "[%s]" letter))
                 (0 ,face prepend)))
  (mapcar* (lambda (var flag) (howm-modify-reminder-types var letter flag))
           '(howm-reminder-types
             howm-schedule-types howm-todo-types
             howm-schedule-menu-types howm-todo-menu-types)
           (list t schedule todo (eq schedule t) (eq todo t))))

(defun howm-modify-reminder-types (var letter flag)
  "Modify variable VAR whose value is \"[...]\".
Example:
(setq foo \"[abc]\")
(howm-modify-reminder-types 'foo \"d\" t)  foo ==> \"[abcd]\"
(howm-modify-reminder-types 'foo \"b\" nil)  foo ==> \"[acd]\"
"
  (let ((val (symbol-value var)))
    (when (not (string-match "^\\[\\(.*\\)\\]$" val))
      (error "Wrong format - %s: %s" var val))
    (let* ((old (match-string-no-properties 1 val))
           (removed (remove (string-to-char letter) old))
           (new (if flag
                    ;; This order is important when val is "[-+~!.]".
                    (concat removed letter)
                  removed)))
      (set var (format "[%s]" new)))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; provide

(provide 'howm-reminder)

;;; howm-reminder.el ends here
